==> allensdk-2.8.0-py3-none-any.whl: allensdk/brain_observatory/ecephys/current_source_density/_current_source_density.py <==
import logging

import numpy as np
import pandas as pd

from typing import Callable, List, Optional, Tuple

from ._interpolation_utils import regular_grid_extractor_factory


def extract_trial_windows(
    stimulus_table: pd.DataFrame, stimulus_name: str,
    time_step: float, pre_stimulus_time: float, post_stimulus_time: float,
    num_trials: Optional[int] = None,
    stimulus_index: Optional[int] = None,
    name_field: str = 'stimulus_name', index_field: str = 'stimulus_index',
    start_field: str = 'Start', end_field: str = 'End'
) -> Tuple[List[np.ndarray], np.ndarray]:
    '''Obtains time interval surrounding stimulus sweep onsets

    Parameters
    ----------
    stimulus_table : pandas.DataFrame
        Each row is a stimulus sweep. Columns report stimulus name and
        parameters for that sweep, as well as its start and end times.
    stimulus_name : str
        Obtain sweeps from stimuli with this name
        (identifies the kind of stimulus presented).
    stimulus_index : Optional[int], optional
        Obtain sweeps from stimuli with this index
        (used to disambiguate presentations of stimuli with the same name),
        by default None.
    time_step : float
        Specifies the step of the resulting temporal domain (seconds).
    pre_stimulus_time : float
        How far before stimulus onset to begin the temporal domain (seconds).
    post_stimulus_time : float
        How far after stimulus onset to end the temporal domain
        (exclusive, seconds).
    num_trials : Optional[int], optional
        A window will be computed for this many sweeps, by default None
    name_field : str, optional
        Column from which to extract stimulus name, by default 'stimulus_name'
    index_field : str, optional
        Column from which to extract stimulus index,
        by default 'stimulus_index'
    start_field : str, optional
        Column from which to extract start times, by default 'Start'
    end_field : str, optional
        Column from which to extract end times, by default 'End'

    Returns
    -------
    Tuple[trial_windows, relative_times]
        trial_windows : List[numpy.ndarray]
            For each trial, an array of timestamps surrounding that
            trial's onset.
        relative_times : numpy.ndarray
            The basic time domain, centered on 0.
    '''

    if stimulus_index is None:
        stimulus_index = np.amin(stimulus_table[stimulus_table[name_field]
                                 == stimulus_name][index_field].values)

    stimulus_name_mask = (stimulus_table[name_field] == stimulus_name)
    stimulus_index_mask = (stimulus_table[index_field] == stimulus_index)
    trials = stimulus_table[stimulus_name_mask & stimulus_index_mask]

    if num_trials is not None:
        trials = trials.iloc[:num_trials, :]
    trials = trials.to_dict('record')

    relative_times = np.arange(-pre_stimulus_time,
                               post_stimulus_time,
                               time_step)
    trial_windows = [relative_times + trial[start_field] for trial in trials]

    msg = 'calculated relative timestamps: {} ({} timestamps per trial)'
    logging.info(msg.format(relative_times, len(relative_times)))
    msg = 'setup {} trial windows spanning {} to {}'
    logging.info(msg.format(len(trial_windows),
                            trial_windows[0][0],
                            trial_windows[-1][-1]))
    return (trial_windows, relative_times)


def accumulate_lfp_data(timestamps: np.ndarray, lfp_raw: np.ndarray,
                        lfp_channels: np.ndarray,
                        trial_windows: List[np.ndarray],
                        volts_per_bit: float = 1.0,
                        extractor_factory: Callable = (
                            regular_grid_extractor_factory)
                        ) -> np.ndarray:
    ''' Extracts slices of LFP data at defined channels and times.

    Parameters
    ----------
    timestamps : numpy.ndarray
        Associates LFP sample indices with times in seconds.
    lfp_raw : numpy.ndarray
        Dimensions are samples X channels.
    lfp_channels : numpy.ndarray
        Indices of channels to be used in accumulation
    trial_windows : List[numpy.ndarray]
        Each window is a list of times from which LFP data will be extracted.
    volts_per_bit: float, optional
        Scaling factor for raw integers into microvolts, defaults to 1.0
        (no conversion)
    extractor_factory: Callable
        The LFP extractor function to use, defaults to
        regular_grid_extractor_factory

    Returns
    -------
    accumulated : numpy.ndarray
        Extracted data. Dimensions are trials X channels X samples

    '''

    num_samples = min(len(tw) for tw in trial_windows)
    num_trials = len(trial_windows)
    num_channels = len(lfp_channels)

    accumulated = np.zeros((num_trials, num_channels, num_samples),
                           dtype=lfp_raw.dtype)

    for channel_idx, chan in enumerate(lfp_channels):
        logging.info('extracting lfp for channel {}'.format(chan))
        extractor = extractor_factory(timestamps, lfp_raw, chan)

        for trial_index, trial_window in enumerate(trial_windows):
            current = extractor(trial_window)[:num_samples]

            if np.issubdtype(accumulated.dtype, np.integer):
                current = np.around(current).astype(accumulated.dtype)
            accumulated[trial_index, channel_idx, :] = current

    msg = 'extracted lfp data for {} trials, {} channels, and {} samples'
    logging.info(msg.format(*accumulated.shape))
    return accumulated * volts_per_bit


def compute_csd(trial_mean_lfp: np.ndarray,
                spacing: float) -> Tuple[np.ndarray, np.ndarray]:
    '''Compute current source density for real or virtual channels from
    a neuropixels probe.

    Compute a second spatial derivative along the probe length
    as a 1D approximation of the Laplacian, after Pitts (1952).

    Parameters
    ----------
    trial_mean_lfp: numpy.ndarray
        LFP traces surrounding presentation of a common stimulus that
        have been averaged over trials. Dimensions are channels X time samples.
    spacing : float
        Distance between channels, in millimeters. This spacing may be
        physical distances between channels or a virtual distance if channels
        have been interpolated to new virtual positions.

    Returns
    -------
    Tuple[csd, csd_channels]:
        csd : numpy.ndarray
            Current source density. Dimensions are channels X time samples.
        csd_channels: numpy.ndarray
            Array of channel indices for CSD.
    '''

    # Need to pad lfp channels for Laplacian approx.
    padded_lfp = np.pad(trial_mean_lfp,
                        pad_width=((1, 1), (0, 0)),
                        mode='edge')

    csd = (1 / (spacing ** 2)) * (padded_lfp[2:, :]
                                  - (2 * padded_lfp[1:-1, :])
                                  + padded_lfp[:-2, :])

    csd_channels = np.arange(0, trial_mean_lfp.shape[0])

    return (csd, csd_channels)

==> allensdk-2.13.6-py3-none-any.whl: allensdk/brain_observatory/ecephys/current_source_density/_current_source_density.py <==
import logging

import numpy as np
import pandas as pd

from typing import Callable, List, Optional, Tuple

from ._interpolation_utils import regular_grid_extractor_factory


def extract_trial_windows(
    stimulus_table: pd.DataFrame,
    stimulus_name: str,
    time_step: float,
    pre_stimulus_time: float,
    post_stimulus_time: float,
    num_trials: Optional[int] = None,
    stimulus_index: Optional[int] = None,
    name_field: str = 'stimulus_name',
    index_field: str = 'stimulus_index',
    start_field: str = 'Start',
    end_field: str = 'End'
) -> Tuple[List[np.ndarray], np.ndarray]:
    '''Obtains time interval surrounding stimulus sweep onsets

    Parameters
    ----------
    stimulus_table : pandas.DataFrame
        Each row is a stimulus sweep. Columns report stimulus name and
        parameters for that sweep, as well as its start and end times.
    stimulus_name : str
        Obtain sweeps from stimuli with this name
        (identifies the kind of stimulus presented).
    stimulus_index : Optional[int], optional
        Obtain sweeps from stimuli with this index
        (used to disambiguate presentations of stimuli with the same name),
        by default None.
    time_step : float
        Specifies the step of the resulting temporal domain (seconds).
    pre_stimulus_time : float
        How far before stimulus onset to begin the temporal domain (seconds).
    post_stimulus_time : float
        How far after stimulus onset to end the temporal domain
        (exclusive, seconds).
    num_trials : Optional[int], optional
        A window will be computed for this many sweeps, by default None
    name_field : str, optional
        Column from which to extract stimulus name, by default 'stimulus_name'
    index_field : str, optional
        Column from which to extract stimulus index,
        by default 'stimulus_index'
    start_field : str, optional
        Column from which to extract start times, by default 'Start'
    end_field : str, optional
        Column from which to extract end times, by default 'End'

    Returns
    -------
    Tuple[trial_windows, relative_times]
        trial_windows : List[numpy.ndarray]
            For each trial, an array of timestamps surrounding that
            trial's onset.
        relative_times : numpy.ndarray
            The basic time domain, centered on 0.
    '''

    if stimulus_index is None:
        stimulus_index = np.amin(stimulus_table[stimulus_table[name_field]
                                 == stimulus_name][index_field].values)

    stimulus_name_mask = (stimulus_table[name_field] == stimulus_name)
    stimulus_index_mask = (stimulus_table[index_field] == stimulus_index)
    trials = stimulus_table[stimulus_name_mask & stimulus_index_mask]

    if num_trials is not None:
        trials = trials.iloc[:num_trials, :]
    trials = trials.to_dict('record')

    relative_times = np.arange(-pre_stimulus_time,
                               post_stimulus_time,
                               time_step)
    trial_windows = [relative_times + trial[start_field] for trial in trials]

    msg = 'calculated relative timestamps: {} ({} timestamps per trial)'
    logging.info(msg.format(relative_times, len(relative_times)))
    msg = 'setup {} trial windows spanning {} to {}'
    logging.info(msg.format(len(trial_windows),
                            trial_windows[0][0],
                            trial_windows[-1][-1]))
    return (trial_windows, relative_times)


def accumulate_lfp_data(timestamps: np.ndarray, lfp_raw: np.ndarray,
                        lfp_channels: np.ndarray,
                        trial_windows: List[np.ndarray],
                        volts_per_bit: float = 1.0,
                        extractor_factory: Callable = (
                            regular_grid_extractor_factory)
                        ) -> np.ndarray:
    ''' Extracts slices of LFP data at defined channels and times.

    Parameters
    ----------
    timestamps : numpy.ndarray
        Associates LFP sample indices with times in seconds.
    lfp_raw : numpy.ndarray
        Dimensions are samples X channels.
    lfp_channels : numpy.ndarray
        Indices of channels to be used in accumulation
    trial_windows : List[numpy.ndarray]
        Each window is a list of times from which LFP data will be extracted.
    volts_per_bit: float, optional
        Scaling factor for raw integers into microvolts, defaults to 1.0
        (no conversion)
    extractor_factory: Callable
        The LFP extractor function to use, defaults to
        regular_grid_extractor_factory

    Returns
    -------
    accumulated : numpy.ndarray
        Extracted data. Dimensions are trials X channels X samples

    '''

    num_samples = min(len(tw) for tw in trial_windows)
    num_trials = len(trial_windows)
    num_channels = len(lfp_channels)

    accumulated = np.zeros((num_trials, num_channels, num_samples),
                           dtype=lfp_raw.dtype)

    for channel_idx, chan in enumerate(lfp_channels):
        logging.info('extracting lfp for channel {}'.format(chan))
        extractor = extractor_factory(timestamps, lfp_raw, chan)

        for trial_index, trial_window in enumerate(trial_windows):
            current = extractor(trial_window)[:num_samples]

            if np.issubdtype(accumulated.dtype, np.integer):
                current = np.around(current).astype(accumulated.dtype)
            accumulated[trial_index, channel_idx, :] = current

    msg = 'extracted lfp data for {} trials, {} channels, and {} samples'
    logging.info(msg.format(*accumulated.shape))
    return accumulated * volts_per_bit


def compute_csd(trial_mean_lfp: np.ndarray,
                spacing: float) -> Tuple[np.ndarray, np.ndarray]:
    '''Compute current source density for real or virtual channels from
    a neuropixels probe.

    Compute a second spatial derivative along the probe length
    as a 1D approximation of the Laplacian, after Pitts (1952).

    Parameters
    ----------
    trial_mean_lfp: numpy.ndarray
        LFP traces surrounding presentation of a common stimulus that
        have been averaged over trials. Dimensions are channels X time samples.
    spacing : float
        Distance between channels, in millimeters. This spacing may be
        physical distances between channels or a virtual distance if channels
        have been interpolated to new virtual positions.

    Returns
    -------
    Tuple[csd, csd_channels]:
        csd : numpy.ndarray
            Current source density. Dimensions are channels X time samples.
        csd_channels: numpy.ndarray
            Array of channel indices for CSD.
    '''

    # Need to pad lfp channels for Laplacian approx.
    padded_lfp = np.pad(trial_mean_lfp,
                        pad_width=((1, 1), (0, 0)),
                        mode='edge')

    csd = (1 / (spacing ** 2)) * (padded_lfp[2:, :]
                                  - (2 * padded_lfp[1:-1, :])
                                  + padded_lfp[:-2, :])

    csd_channels = np.arange(0, trial_mean_lfp.shape[0])

    return (csd, csd_channels)
